@@ -30,20 +30,25 @@ import subprocess
 import time
 from itertools import permutations
 
-from mimetypes import guess_type
+import mimetypes
+import zipfile
 
 import gi
 gi.require_version('GLib', '2.0')  # noqa
 from gi.repository import GLib
 
 try:
-    from zeitgeist.client import ZeitgeistDBusInterface  # pylint: disable=E0401
-    from zeitgeist.datamodel import Event, TimeRange  # pylint: disable=E0401
-    from zeitgeist import datamodel  # pylint: disable=E0401
-    IFACE = ZeitgeistDBusInterface()
-    ZEITGEIST_SUPPORT = True
+    gi.require_version('Zeitgeist', '2.0')
+    from gi.repository import Zeitgeist
+    log_trial = Zeitgeist.Log.get_default()
+    if log_trial.datapath() is None:
+        ZEITGEIST_SUPPORT = False
+    else:
+        ZEITGEIST_SUPPORT = True
 except ImportError:
     ZEITGEIST_SUPPORT = False
+except ValueError:
+    ZEITGEIST_SUPPORT = False
 
 LOGGER = logging.getLogger('catfish_search')
 ENGINE_COUNT = 0
@@ -145,7 +150,7 @@ class CatfishSearchEngine:
         """Add a CatfishSearchMethod the the engine's search backends."""
         self.methods.append(method_class())
 
-    def run(self, keywords, path, limit=-1, regex=False):  # noqa
+    def run(self, keywords, path, search_zips, limit=-1, regex=False):  # noqa
         """Run the CatfishSearchEngine.
 
         Each method is run sequentially in the order they are added.
@@ -171,6 +176,11 @@ class CatfishSearchEngine:
             else:
                 keys.append(key)
 
+        # path may be empty when some "extension" scheme is used in URL
+        if not path:
+            self.stop()
+            return
+
         # For simplicity, make sure the path contains a trailing '/'
         if not path.endswith('/'):
             path += '/'
@@ -190,7 +200,7 @@ class CatfishSearchEngine:
             LOGGER.debug(
                 "[%i] Starting search method: %s",
                 self.engine_id, method.method_name)
-            for filename in method.run(keywords, path, regex,
+            for filename in method.run(keywords, path, search_zips, regex,
                                        self.exclude_paths):
                 if isinstance(filename, str) and path in filename:
                     found_bad = False
@@ -204,7 +214,13 @@ class CatfishSearchEngine:
                         yield True
                         continue
 
-                    if method.method_name == 'fulltext' or  \
+                    if search_zips and method.method_name == 'walk':
+                        if os.path.isfile(filename) and \
+                                zipfile.is_zipfile(filename):
+                            yield filename
+                            continue
+
+                    if method.method_name == 'fulltext' or \
                             all(key in
                                 os.path.basename(filename).lower()
                                 for key in keys):
@@ -253,6 +269,33 @@ class CatfishSearchEngine:
                 yield False
         self.stop()
 
+    def search_zip(self, fullpath, keywords, search_exact):
+        keyword_list = get_keyword_list(keywords)
+        with zipfile.ZipFile(fullpath, 'r') as z:
+            for member in z.infolist():
+                for method in self.methods:
+                    if method.method_name == 'walk':
+                        if self.search_filenames(
+                                member.filename, keywords, search_exact):
+                            yield (member.filename, member.file_size, member.date_time)
+                    if method.method_name == 'fulltext':
+                        if method.search_zip(z, member, keyword_list):
+                            yield (member.filename, member.file_size, member.date_time)
+
+    def search_filenames(self, filename, keywords, search_exact):
+        keywords = get_keyword_list(keywords)
+        fname = os.path.basename(filename.rstrip("/"))
+        if search_exact:
+            if " ".join(keywords) in fname.lower():
+                return True
+        else:
+            match_list = set()
+            for kword in keywords:
+                if kword in fname.lower():
+                    match_list.add(kword)
+                if len(set(keywords)) == len(match_list):
+                    return True
+
     def set_exact(self, exact):
         """Set method for exact"""
         # Only for fulltext engine
@@ -277,10 +320,14 @@ class CatfishSearchMethod:
         """Base CatfishSearchMethod Initializer."""
         self.method_name = method_name
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):  # pylint: disable=W0613
+    def run(self, keywords, path, search_zips, regex=False, exclude_paths=[]):  # pylint: disable=W0613
         """Base CatfishSearchMethod run method."""
         return NotImplemented
 
+    def search_zip(self, z, member, keywords):
+        """Base CatfishSearchMethod search_zip method."""
+        return False
+
     def stop(self):
         """Base CatfishSearchMethod stop method."""
         return NotImplemented
@@ -337,7 +384,6 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
                 # Link: ~/.wine/dosdevices/z:/ -> /
                 if realroot.startswith(realpath):
                     continue
-                processed_links.append(realpath)
             if os.path.basename(path).startswith("."):
                 if islink:
                     dotlinks.append(path)
@@ -352,7 +398,7 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
         dirlist = xdgdirs + notdotdirs + notdotlinks + dotdirs + dotlinks
         return dirlist
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):
+    def run(self, keywords, path, search_zips, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
@@ -373,15 +419,14 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
 
         # Grab the special directory list to get them precedence
         xdgdirlist = [
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DESKTOP),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOCUMENTS),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOWNLOAD),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES),
-            GLib.get_user_special_dir(
-                GLib.USER_DIRECTORY_PUBLIC_SHARE),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_TEMPLATES),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_VIDEOS),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOCUMENTS),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOWNLOAD),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_MUSIC),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PUBLIC_SHARE),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_TEMPLATES),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_VIDEOS),
         ]
 
         for root, dirs, files in os.walk(top=path, topdown=True,
@@ -408,11 +453,24 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
             # Check paths in the second and deeper levels of the selected
             # directory
             for path in paths:
-                if any(keyword in path.lower() for keyword in keywords):
-                    yield os.path.join(root, path)
+                fullpath = os.path.join(root, path)
+                if self.search_path(path, keywords):
+                    yield fullpath
+                if search_zips and \
+                   os.path.isfile(fullpath) and zipfile.is_zipfile(fullpath):
+                    yield fullpath
             yield True
         yield False
 
+    def search_path(self, path, keywords):
+        if any(keyword in path.lower() for keyword in keywords):
+            return True
+
+    def search_zip(self, z, member, keywords):
+        if member.is_dir():
+            return self.search_path(member.filename, keywords)
+        return self.search_path(os.path.basename(member.filename), keywords)
+
     def stop(self):
         """Stop the running search method."""
         self.running = False
@@ -435,7 +493,86 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         self.running = False
         self.exact = False
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):  # noqa
+    def check_charset(self, root, filename):
+        """Checks file for BOM to detect UTF encodings.
+        Also checks file for null byte. Most files containing
+        a null byte will be binary. UTF-16 and UTF-32 without
+        BOM can contain null bytes, support for that can be
+        added by checking the line ending."""
+        file = open(os.path.join(root, filename), 'rb')
+        try:
+            data = file.read(64).lower()
+
+            if data.startswith(b'\x00\x00\xfe\xff'):
+                return 'utf-32-be'
+
+            elif data.startswith(b'\xff\xfe\x00\x00'):
+                return 'utf-32-le'
+
+            elif data.startswith(b'\x2b\x2f\x76'):
+                return 'utf-7'
+
+            elif data.startswith(b'\xfe\xff'):
+                return 'utf-16-be'
+
+            elif data.startswith(b'\xff\xfe'):
+                return 'utf-16-le'
+
+            elif b'\0' in data:
+                return 'binary'
+
+            else:
+                return 'utf-8'
+
+        finally:
+            file.close()
+
+    def is_txt(self, filename):
+        """Checks if text mimetype."""
+        mime = str(mimetypes.guess_type(filename)[0])
+        text_list = ('ardour', 'audacity', 'desktop', 'document',
+                     'fontforge', 'java', 'json', 'm4', 'mbox',
+                     'message', 'mimearchive', 'msg', 'none', 'perl',
+                     'pgp-keys', 'php', 'postscript', 'rtf',
+                     'ruby', 'shellscript', 'spreadsheet', 'sql',
+                     'subrip', 'text', 'troff', 'url', 'winhlp',
+                     'x-bittorent', 'x-cue', 'x-extension-cfg',
+                     'x-glade', 'x-mpegurl', 'x-sami', 'x-theme',
+                     'x-trash', 'xml', 'xpixmap', 'yaml')
+        for filetype in text_list:
+            if filetype in mime.lower():
+                return True
+
+    def search_pdf(self, fullpath, keywords):
+        command = ['pdftotext', '-q', '-nopgbrk', fullpath, '-']
+        with subprocess.Popen(command, stdout=subprocess.PIPE, text=True) as pdf:
+            if self.search_text(pdf.stdout, keywords):
+                return True
+
+    def search_text(self, lines, keywords):
+        if self.exact:
+            for line in lines:
+                if " ".join(keywords) in line.lower():
+                    return True
+        else:
+            match_list = set()
+            for line in lines:
+                for kword in keywords:
+                    if kword in line.lower():
+                        match_list.add(kword)
+                if len(set(keywords)) == len(match_list):
+                    return True
+
+    def search_zip(self, z, member, keywords):
+        with z.open(member) as f:
+            lines = io.TextIOWrapper(f, encoding='utf-8')
+            try:
+                if self.search_text(lines, keywords):
+                    return True
+            except UnicodeError:
+                return False
+
+    def run(self, keywords, path, search_zips, regex=False, exclude_paths=[]):  # noqa
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
@@ -457,50 +594,50 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         for root, dirs, files in os.walk(path):  # pylint: disable=W0612
             if self.force_stop:
                 break
+            # Don't search user excluded directories.
+            if root.startswith(tuple(exclude_paths)):
+                dirs[:] = []
+                continue
 
             for filename in files:
-                if self.force_stop:
-                    break
+                try:
+                    fullpath = os.path.join(root, filename)
 
-                # If the filetype is known to not be text, move along.
-                mime = guess_type(filename)[0]
-                if not mime or 'text' in mime:
-                    try:
-                        opened = open(os.path.join(root, filename), 'r')
+                    # Skip if special file.
+                    if not os.path.isfile(fullpath):
+                        continue
+                    if os.path.getsize(fullpath) == 0:
+                        continue
+                    if fullpath.lower().endswith('.pdf'):
+                        if self.search_pdf(fullpath, keywords):
+                            yield fullpath
+                    if zipfile.is_zipfile(fullpath):
+                        yield fullpath
+                    # Skip if not text file.
+                    if not self.is_txt(filename):
+                        continue
+                    # Check character encoding, skip if binary.
+                    charset = self.check_charset(root, filename)
+                    if charset == 'binary':
+                        continue
 
-                        find_keywords = find_keywords_backup
-
-                        # Check each line.  If a keyword is found, yield.
-                        try:
-                            for line in opened:
-                                if self.force_stop:
-                                    break
-
-                                if self.exact:
-                                    if " ".join(keywords) in line:
-                                        yield os.path.join(root, filename)
-                                        break
-                                else:
-                                    if any(keyword in line.lower()
-                                           for keyword in keywords):
-                                        found_keywords = []
-                                        for find_keyword in find_keywords:
-                                            if find_keyword in line.lower():
-                                                found_keywords.append(
-                                                    find_keyword)
-                                        for found_keyword in found_keywords:
-                                            find_keywords.remove(found_keyword)
-
-                                        if len(find_keywords) == 0:
-                                            yield os.path.join(root, filename)
-                                            break
-                        except UnicodeDecodeError:
-                            pass
-
-                        opened.close()
-                    except IOError:
-                        pass
-                yield True
+                    # Check each line. If a keyword is found, yield.
+                    open_file = open(fullpath, 'r', encoding=charset)
+                    with open_file as file_text:
+                        if self.search_text(file_text, keywords):
+                            yield fullpath
+                # Skips on errors, move on to next in list.
+                except UnicodeDecodeError:
+                    continue
+                except UnicodeError:
+                    continue
+                except FileNotFoundError:
+                    continue
+                except PermissionError:
+                    continue
+                except OSError:
+                    continue
+            yield True
         yield False
         self.force_stop = False
         self.running = False
@@ -523,30 +660,47 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
         """Initialize the Zeitgeist SearchMethod."""
         CatfishSearchMethod.__init__(self, "zeitgeist")
         self.stop_search = False
+        self.events = []
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):
+    def result_callback(self, log, result, data):
+        events = log.find_events_finish(result)
+        for i in range(events.size()):  # cannot iterate events directly
+            event = events.next_value()
+            if event is None:
+                continue
+            self.events.append(event)
+        self.stop_search = True
+
+    def run(self, keywords, path, search_zips, regex=False, exclude_paths=[]):
         """Run the Zeitgeist SearchMethod."""
+        keywords = " ".join(keywords)
         self.stop_search = False
-        event_template = Event()
-        time_range = TimeRange.from_seconds_ago(60 * 3600 * 24)
-        # 60 days at most
+        event_template = Zeitgeist.Event()
 
-        results = IFACE.FindEvents(
-            time_range,  # (min_timestamp, max_timestamp) in milliseconds
+        # 60 days at most
+        end = int(time.time() * 1000)
+        start = end - (60 * 3600 * 24 * 1000)
+        time_range = Zeitgeist.TimeRange.new(start, end)
+
+        self.events = []
+        log = Zeitgeist.Log.get_default()
+        log.find_events(
+            time_range,
             [event_template, ],
-            datamodel.StorageState.Any,
+            Zeitgeist.StorageState.ANY,
             1000,
-            datamodel.ResultType.MostRecentSubjects
+            Zeitgeist.ResultType.MOST_RECENT_SUBJECTS,
+            None, self.result_callback, None
         )
 
-        results = (datamodel.Event(result) for result in results)
+        while self.stop_search == False:
+            yield False
+
         uniques = []
 
-        for event in results:
-            if self.stop_search:
-                break
+        for event in self.events:
             for subject in event.get_subjects():
-                uri = str(subject.uri)
+                uri = subject.get_uri()
                 if uri.startswith('file://'):
                     fullname = str(uri[7:])
                     filepath, filename = os.path.split(fullname)
@@ -555,7 +709,6 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
                             path in filepath:
                         uniques.append(uri)
                         yield fullname
-        self.stop_search = True
 
     def stop(self):
         """Stop the Zeitgeist SearchMethod."""
@@ -582,7 +735,7 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
         """Base assemble_query method."""
         return False
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):
+    def run(self, keywords, path, search_zips, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.
 
         This function returns the process.stdout generator and will yield files
